"""Module with utility functions for exception handling"""
import logging
import traceback
from socket import gethostname

from pydantic import BaseModel

from .general_notifier import GeneralNotifier


class ExceptionNotifier(BaseModel):
    is_exception_in_subject: bool = True
    salutation: str = 'Hi,'
    subject: str = ''

    class Config:
        arbitrary_types_allowed = True

    def notify(self, exception: Exception):
        """Sends a notification of the given exception"""
        exception_name = exception.__class__.__name__
        exception_traceback = traceback.format_exc()
        extended_subject = f'{self.subject}'

        if self.is_exception_in_subject:
            extended_subject = f'{extended_subject} {exception_name}'

        extended_subject = f'{extended_subject} on Host {gethostname()}'

        extended_body = f'''\
        {self.salutation}
        <br /><br />
        <strong>{exception_name}</strong>
        <br />
        <div style="background-color: #ebeff5; padding: 1rem; margin: 1rem;">
            <div style="white-space: pre-line;">
                {exception_traceback}
            </div>
        </div>
        '''

        logging.error(f'{exception_name}\n{exception_traceback}')
        GeneralNotifier.send(subject=extended_subject, body=extended_body, priority=1)



"""
Utility Functions for sending email
"""
import logging
import os
import smtplib
import ssl
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from typing import Optional, List


class GeneralNotifier:
    host: str = os.getenv('MAIL_SERVER_HOST', 'smtp.office365.com')
    port: int = int(os.getenv('MAIL_SERVER_PORT', '587'))
    username: str = os.getenv('DEFAULT_SENDER_EMAIL_ADDRESS', '')
    password: str = os.getenv('DEFAULT_SENDER_EMAIL_PASSWORD', '')
    subject_prefix: str = os.getenv('DEFAULT_EMAIL_SUBJECT_PREFIX', '')
    email_signature: str = os.getenv('DEFAULT_EMAIL_SIGNATURE', '')
    default_recipients: List[str] = [
        email.strip() for email in os.getenv('DEFAULT_EMAIL_RECIPIENTS', '').split(',')
    ]

    @classmethod
    def _start_tls(cls, server: smtplib.SMTP):
        """Starts an SSL connection attempt"""
        try:
            context = ssl.create_default_context()
            server.ehlo()
            server.starttls(context=context)
            server.ehlo()
        except smtplib.SMTPNotSupportedError as exp:
            logging.info('Email SMTP server does not support SSL')

    @classmethod
    def _login(cls) -> Optional[smtplib.SMTP]:
        """Logs into the SMTP server and returns the server"""
        server = smtplib.SMTP(cls.host, cls.port)
        cls._start_tls(server=server)

        try:
            server.login(cls.username, cls.password)
        except smtplib.SMTPNotSupportedError as exp:
            logging.info('Email SMTP server does not support authentication')

        return server

    @classmethod
    def _construct_message(cls, subject: str, body: str, recipients: List[str], priority: int = 3) -> str:
        """Constructs a message to be sent via email"""
        message = MIMEMultipart("alternative")
        message["Subject"] = f"{cls.subject_prefix}{subject}"
        message["From"] = cls.username
        message["To"] = ','.join(recipients)
        message['X-Priority'] = f'{priority}'

        if priority == 1:
            message['X-MSMail-Priority'] = 'High'
            message['Importance'] = 'High'

        text = f"""
        {body}

        {cls.email_signature}
        """

        html = f"""
        <html>
            <body>
                <div>{body}</div>
                <br />
                <div>{cls.email_signature}</div>
            </body>
        </html>
        """
        # Turn these into plain/html MIMEText objects
        main_message_part = MIMEText(html, 'html')
        backup_message_part = MIMEText(text, 'text')
        # Attach the MIMEText objects. Email client attempts the last part first
        message.attach(backup_message_part)
        message.attach(main_message_part)

        return message.as_string()

    @classmethod
    def send(cls, subject: str, body: str, recipients: Optional[List[str]] = None, priority: int = 3):
        """Sends the email"""
        server: Optional[smtplib.SMTP] = None
        return_value = None

        try:
            server = cls._login()
            recipient_emails = cls.default_recipients if recipients is None else recipients
            message = cls._construct_message(subject=subject, body=body, recipients=recipient_emails)
            return_value = server.sendmail(from_addr=cls.username, to_addrs=recipient_emails, msg=message)
        except Exception as exp:
            logging.error(exp)
        finally:
            if server is not None:
                server.quit()

        return return_value



"""Tests for the GeneralNotifier class"""
# import logging
import logging
import os
from socket import gethostname
from traceback import format_exc
from unittest import TestCase, main
from unittest.mock import Mock, patch

from dotenv import load_dotenv

load_dotenv(os.path.join(os.path.dirname(os.path.dirname(__file__)), '.env'))

from test.mock_smtp import MockSMTPServer
from email_notifier import GeneralNotifier, ExceptionNotifier


class TestNotifier(TestCase):
    """Test class for the GeneralNotifier and the ExceptionNotifier classes"""
    mock_server: MockSMTPServer

    @classmethod
    def setUpClass(cls) -> None:
        cls.mock_server = MockSMTPServer(
            hostname=os.getenv('MAIL_SERVER_HOST', '127.0.0.1'),
            port=int(os.getenv('MAIL_SERVER_PORT', 10025)))

    def setUp(self) -> None:
        """Initialize some variables"""
        self.subject_prefix = os.getenv('DEFAULT_EMAIL_SUBJECT_PREFIX', '')
        self.notifier = GeneralNotifier()
        self.subject = 'Hello World'
        self.salutation = 'Hi there'
        self.body = 'Hi there'
        self.sender = os.getenv('DEFAULT_SENDER_EMAIL_ADDRESS')
        self.recipients = [
            email.strip() for email in os.getenv('NON_DEFAULT_EMAIL_RECIPIENTS', '').split(',')]
        self.default_recipients = [
            email.strip() for email in os.getenv('DEFAULT_EMAIL_RECIPIENTS', '').split(',')]

    def test_general_notifier_send(self):
        """Should send message to a given set of recipients"""
        self.assertEqual(self.mock_server.received_messages_count(), 0)
        self.notifier.send(subject=self.subject, body=self.body, recipients=self.recipients)
        self.assertGreater(self.mock_server.received_messages_count(), 0)

        self.assertTrue(all([self.mock_server.has_message_for(recipient) for recipient in self.recipients]))
        self.assertTrue(self.mock_server.received_message_from(self.sender))
        self.assertTrue(self.mock_server.received_message_matching(self.body))
        self.assertTrue(self.mock_server.received_message_matching(f'Subject: {self.subject_prefix}{self.subject}'))

    def test_general_notifier_send_default_recipients(self):
        """If no recipients are provided, it sends to those in the environment"""
        self.assertEqual(self.mock_server.received_messages_count(), 0)
        self.notifier.send(subject=self.subject, body=self.body)
        self.assertGreater(self.mock_server.received_messages_count(), 0)

        self.assertTrue(all([self.mock_server.has_message_for(recipient) for recipient in self.default_recipients]))
        self.assertTrue(self.mock_server.received_message_from(self.sender))
        self.assertTrue(self.mock_server.received_message_matching(self.body))
        self.assertTrue(self.mock_server.received_message_matching(f'Subject: {self.subject_prefix}{self.subject}'))

    def test_exception_notifier_notify_email(self):
        """Should send email of the exception to the recipients set in the environment"""
        notifier = ExceptionNotifier(subject=self.subject, salutation=self.salutation, is_exception_in_subject=False)

        def trial_function():
            try:
                raise ValueError('Dummy error')
            except ValueError as exp:
                notifier.notify(exp)

        trial_function()

        host = gethostname()
        subject = f'{self.subject_prefix}{self.subject} on Host {host}'
        self.assertEqual(self.mock_server.received_messages_count(), 1)
        self.assertTrue(all([self.mock_server.has_message_for(recipient) for recipient in self.default_recipients]))
        self.assertTrue(self.mock_server.received_message_from(self.sender))
        self.assertTrue(self.mock_server.received_message_matching(self.salutation))
        self.assertTrue(self.mock_server.received_message_matching(f'Subject: {subject}'))
        self.assertTrue(self.mock_server.received_message_matching(r'Traceback \(most recent call last\)'))

    def test_exception_notifier_default_notify_email(self):
        """Should send email of the exception to the recipients set in the environment"""
        notifier = ExceptionNotifier()

        def trial_function():
            try:
                raise ValueError('Dummy error')
            except ValueError as exp:
                notifier.notify(exp)

        trial_function()

        exception_name = 'ValueError'
        host = gethostname()
        subject = f'{self.subject_prefix} {exception_name} on Host {host}'

        self.assertEqual(self.mock_server.received_messages_count(), 1)
        self.assertTrue(all([self.mock_server.has_message_for(recipient) for recipient in self.default_recipients]))
        self.assertTrue(self.mock_server.received_message_from(self.sender))
        self.assertFalse(self.mock_server.received_message_matching(self.salutation))
        self.assertTrue(self.mock_server.received_message_matching(f'Subject: {subject}'))
        self.assertTrue(self.mock_server.received_message_matching(r'Traceback \(most recent call last\)'))

    @patch.object(logging, 'error')
    def test_exception_notifier_notify_log(self, mocked_error_logger: Mock):
        """Should log the exception using the python logger"""
        notifier = ExceptionNotifier()

        def trial_function():
            try:
                raise ValueError('Dummy error')
            except ValueError as exp:
                notifier.notify(exp)
                return format_exc()

        exception_traceback = trial_function()

        exception_name = 'ValueError'
        mocked_error_logger.assert_called_with(f'{exception_name}\n{exception_traceback}')

    def tearDown(self) -> None:
        """Reset the mock SMTP server"""
        self.mock_server.reset()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.mock_server.close()


if __name__ == '__main__':
    main()



import pathlib
from setuptools import setup, find_packages

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
    name="email_notifier",
    version="0.0.1",
    description="A simple email notifier for general messages or exceptions with easy environment configuration",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/sopherapps/email_notifier",
    author="Martin Ahindura",
    author_email="team.sopherapps@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    packages=find_packages(exclude=("test",)),
    include_package_data=True,
    install_requires=["pydantic"],
    entry_points={
    },
)